@@ -2,6 +2,7 @@
 
 use markup5ever::{expanded_name, local_name, namespace_url, ns};
 use std::cell::RefCell;
+use std::convert::TryFrom;
 use std::rc::Rc;
 
 use crate::bbox::BoundingBox;
@@ -9,17 +10,26 @@ use crate::document::{AcquiredNodes, NodeId};
 use crate::drawing_ctx::{DrawingCtx, ViewParams};
 use crate::element::{Draw, Element, ElementResult, SetAttributes};
 use crate::error::*;
-use crate::layout::{self, FontProperties, StackingContext, Stroke};
+use crate::layout::{self, FontProperties, StackingContext, Stroke, TextSpan};
 use crate::length::*;
 use crate::node::{CascadedValues, Node, NodeBorrow};
+use crate::paint_server::PaintSource;
 use crate::parsers::ParseValue;
 use crate::properties::{
-    ComputedValues, Direction, FontStretch, FontStyle, FontVariant, FontWeight, TextAnchor,
-    UnicodeBidi, WritingMode, XmlLang, XmlSpace,
+    ComputedValues, Direction, FontStretch, FontStyle, FontVariant, FontWeight, PaintOrder,
+    TextAnchor, TextRendering, UnicodeBidi, WritingMode, XmlLang, XmlSpace,
 };
+use crate::rect::Rect;
 use crate::space::{xml_space_normalize, NormalizeDefault, XmlSpaceNormalize};
+use crate::transform::Transform;
 use crate::xml::Attributes;
 
+/// The state of a text layout operation.
+struct LayoutContext {
+    /// `writing-mode` property from the `<text>` element.
+    writing_mode: WritingMode,
+}
+
 /// An absolutely-positioned array of `Span`s
 ///
 /// SVG defines a "[text chunk]" to occur when a text-related element
@@ -38,23 +48,21 @@ struct Chunk {
     x: Option<f64>,
     y: Option<f64>,
     spans: Vec<Span>,
-    link: Option<String>,
 }
 
 struct MeasuredChunk {
     values: Rc<ComputedValues>,
     x: Option<f64>,
     y: Option<f64>,
-    advance: (f64, f64),
+    dx: f64,
+    dy: f64,
     spans: Vec<MeasuredSpan>,
-    link: Option<String>,
 }
 
 struct PositionedChunk {
     next_chunk_x: f64,
     next_chunk_y: f64,
     spans: Vec<PositionedSpan>,
-    link: Option<String>,
 }
 
 struct Span {
@@ -63,30 +71,54 @@ struct Span {
     dx: f64,
     dy: f64,
     _depth: usize,
+    link_target: Option<String>,
 }
 
 struct MeasuredSpan {
     values: Rc<ComputedValues>,
     layout: pango::Layout,
-    _layout_size: (f64, f64),
+    layout_size: (f64, f64),
     advance: (f64, f64),
     dx: f64,
     dy: f64,
+    link_target: Option<String>,
 }
 
 struct PositionedSpan {
     layout: pango::Layout,
     values: Rc<ComputedValues>,
     rendered_position: (f64, f64),
+    next_span_position: (f64, f64),
+    link_target: Option<String>,
+}
+
+/// A laid-out and resolved text span.
+///
+/// The only thing not in user-space units are the `stroke_paint` and `fill_paint`.
+///
+/// This is the non-user-space version of `layout::TextSpan`.
+struct LayoutSpan {
+    layout: pango::Layout,
+    gravity: pango::Gravity,
+    bbox: Option<BoundingBox>,
+    is_visible: bool,
+    x: f64,
+    y: f64,
+    paint_order: PaintOrder,
+    stroke: Stroke,
+    stroke_paint: PaintSource,
+    fill_paint: PaintSource,
+    text_rendering: TextRendering,
+    link_target: Option<String>,
+    values: Rc<ComputedValues>,
 }
 
 impl Chunk {
-    fn new(values: &ComputedValues, x: Option<f64>, y: Option<f64>, link: Option<String>) -> Chunk {
+    fn new(values: &ComputedValues, x: Option<f64>, y: Option<f64>) -> Chunk {
         Chunk {
             values: Rc::new(values.clone()),
             x,
             y,
-            link,
             spans: Vec::new(),
         }
     }
@@ -94,125 +126,179 @@ impl Chunk {
 
 impl MeasuredChunk {
     fn from_chunk(
+        layout_context: &LayoutContext,
         chunk: &Chunk,
-        text_writing_mode: WritingMode,
         draw_ctx: &DrawingCtx,
     ) -> MeasuredChunk {
-        let measured_spans: Vec<MeasuredSpan> = chunk
+        let mut measured_spans: Vec<MeasuredSpan> = chunk
             .spans
             .iter()
-            .map(|span| MeasuredSpan::from_span(span, text_writing_mode, draw_ctx))
+            .map(|span| MeasuredSpan::from_span(layout_context, span, draw_ctx))
             .collect();
 
-        let advance = measured_spans.iter().fold((0.0, 0.0), |acc, measured| {
-            (acc.0 + measured.advance.0, acc.1 + measured.advance.1)
-        });
+        // The first span contains the (dx, dy) that will be applied to the whole chunk.
+        // Make them 0 in the span, and extract the values to set them on the chunk.
+        // This is a hack until librsvg adds support for multiple dx/dy values per text/tspan.
+
+        let (chunk_dx, chunk_dy) = if let Some(first) = measured_spans.first_mut() {
+            let dx = first.dx;
+            let dy = first.dy;
+            first.dx = 0.0;
+            first.dy = 0.0;
+            (dx, dy)
+        } else {
+            (0.0, 0.0)
+        };
 
         MeasuredChunk {
             values: chunk.values.clone(),
             x: chunk.x,
             y: chunk.y,
-            advance,
+            dx: chunk_dx,
+            dy: chunk_dy,
             spans: measured_spans,
-            link: chunk.link.clone(),
         }
     }
 }
 
 impl PositionedChunk {
     fn from_measured(
+        layout_context: &LayoutContext,
         measured: &MeasuredChunk,
         view_params: &ViewParams,
-        text_writing_mode: WritingMode,
-        x: f64,
-        y: f64,
+        chunk_x: f64,
+        chunk_y: f64,
     ) -> PositionedChunk {
+        let chunk_direction = measured.values.direction();
+
+        // Position the spans relatively to each other, starting at (0, 0)
+
         let mut positioned = Vec::new();
 
-        // measured.advance is the size of the chunk.  Compute the offsets needed to align
-        // it per the text-anchor property (start, middle, end):
-        let anchor_offset = text_anchor_offset(
-            measured.values.text_anchor(),
-            measured.values.direction(),
-            text_writing_mode,
-            measured.advance,
-        );
+        // Start position of each span; gets advanced as each span is laid out.
+        // This is the text's start position, not the bounding box.
+        let mut x = 0.0;
+        let mut y = 0.0;
 
-        let mut x = x + anchor_offset.0;
-        let mut y = y + anchor_offset.1;
-
-        // Position each span
+        let mut chunk_bounds: Option<Rect> = None;
 
         for mspan in &measured.spans {
             let params = NormalizeParams::new(&mspan.values, view_params);
 
             let layout = mspan.layout.clone();
+            let layout_size = mspan.layout_size;
             let values = mspan.values.clone();
-
-            let baseline = f64::from(layout.baseline()) / f64::from(pango::SCALE);
-            let baseline_shift = values.baseline_shift().0.to_user(&params);
-            let baseline_offset = baseline + baseline_shift;
-
             let dx = mspan.dx;
             let dy = mspan.dy;
+            let advance = mspan.advance;
 
-            let start_pos = match values.direction() {
+            let baseline_offset = compute_baseline_offset(&layout, &values, &params);
+
+            let start_pos = match chunk_direction {
                 Direction::Ltr => (x, y),
-                Direction::Rtl => (x - mspan.advance.0, y),
+                Direction::Rtl => (x - advance.0, y),
             };
 
-            let span_advance = match values.direction() {
-                Direction::Ltr => (mspan.advance.0, mspan.advance.1),
-                Direction::Rtl => (-mspan.advance.0, mspan.advance.1),
+            let span_advance = match chunk_direction {
+                Direction::Ltr => (advance.0, advance.1),
+                Direction::Rtl => (-advance.0, advance.1),
             };
 
-            let rendered_position = if text_writing_mode.is_horizontal() {
+            let rendered_position = if layout_context.writing_mode.is_horizontal() {
                 (start_pos.0 + dx, start_pos.1 - baseline_offset + dy)
             } else {
                 (start_pos.0 + baseline_offset + dx, start_pos.1 + dy)
             };
 
+            let span_bounds =
+                Rect::from_size(layout_size.0, layout_size.1).translate(rendered_position);
+
+            if let Some(bounds) = chunk_bounds {
+                chunk_bounds = Some(bounds.union(&span_bounds));
+            } else {
+                chunk_bounds = Some(span_bounds);
+            }
+
+            x = x + span_advance.0 + dx;
+            y = y + span_advance.1 + dy;
+
             let positioned_span = PositionedSpan {
                 layout,
                 values,
                 rendered_position,
+                next_span_position: (x, y),
+                link_target: mspan.link_target.clone(),
             };
 
             positioned.push(positioned_span);
+        }
 
-            x = x + span_advance.0 + dx;
-            y = y + span_advance.1 + dy;
+        // Compute the offsets needed to align the chunk per the text-anchor property (start, middle, end):
+
+        let anchor_offset = text_anchor_offset(
+            measured.values.text_anchor(),
+            chunk_direction,
+            layout_context.writing_mode,
+            chunk_bounds.unwrap_or_default(),
+        );
+
+        // Apply the text-anchor offset to each individually-positioned span, and compute the
+        // start position of the next chunk.  Also add in the chunk's dx/dy.
+
+        let mut next_chunk_x = chunk_x;
+        let mut next_chunk_y = chunk_y;
+
+        for pspan in &mut positioned {
+            // Add the chunk's position, plus the text-anchor offset, plus the chunk's dx/dy.
+            // This last term is a hack until librsvg adds support for multiple dx/dy values per text/tspan;
+            // see the corresponding part in MeasuredChunk::from_chunk().
+            pspan.rendered_position.0 += chunk_x + anchor_offset.0 + measured.dx;
+            pspan.rendered_position.1 += chunk_y + anchor_offset.1 + measured.dy;
+
+            next_chunk_x = chunk_x + pspan.next_span_position.0 + anchor_offset.0 + measured.dx;
+            next_chunk_y = chunk_y + pspan.next_span_position.1 + anchor_offset.1 + measured.dy;
         }
 
         PositionedChunk {
-            next_chunk_x: x,
-            next_chunk_y: y,
+            next_chunk_x,
+            next_chunk_y,
             spans: positioned,
-            link: measured.link.clone(),
         }
     }
 }
 
+fn compute_baseline_offset(
+    layout: &pango::Layout,
+    values: &ComputedValues,
+    params: &NormalizeParams,
+) -> f64 {
+    let baseline = f64::from(layout.baseline()) / f64::from(pango::SCALE);
+    let baseline_shift = values.baseline_shift().0.to_user(params);
+    baseline + baseline_shift
+}
+
 /// Computes the (x, y) offsets to be applied to spans after applying the text-anchor property (start, middle, end).
 #[rustfmt::skip]
 fn text_anchor_offset(
     anchor: TextAnchor,
     direction: Direction,
-    text_writing_mode: WritingMode,
-    chunk_size: (f64, f64),
+    writing_mode: WritingMode,
+    chunk_bounds: Rect,
 ) -> (f64, f64) {
-    let (w, h) = chunk_size;
+    let (w, h) = (chunk_bounds.width(), chunk_bounds.height());
 
-    if text_writing_mode.is_horizontal() {
+    let x0 = chunk_bounds.x0;
+
+    if writing_mode.is_horizontal() {
         match (anchor, direction) {
-            (TextAnchor::Start,  Direction::Ltr) => (0.0, 0.0),
-            (TextAnchor::Start,  Direction::Rtl) => (0.0, 0.0),
+            (TextAnchor::Start,  Direction::Ltr) => (-x0, 0.0),
+            (TextAnchor::Start,  Direction::Rtl) => (-x0 - w, 0.0),
 
-            (TextAnchor::Middle, Direction::Ltr) => (-w / 2.0, 0.0),
-            (TextAnchor::Middle, Direction::Rtl) => (w / 2.0, 0.0),
+            (TextAnchor::Middle, Direction::Ltr) => (-x0 - w / 2.0, 0.0),
+            (TextAnchor::Middle, Direction::Rtl) => (-x0 - w / 2.0, 0.0),
 
-            (TextAnchor::End,    Direction::Ltr) => (-w, 0.0),
-            (TextAnchor::End,    Direction::Rtl) => (w, 0.0),
+            (TextAnchor::End,    Direction::Ltr) => (-x0 - w, 0.0),
+            (TextAnchor::End,    Direction::Rtl) => (-x0, 0.0),
         }
     } else {
         // FIXME: we don't deal with text direction for vertical text yet.
@@ -225,21 +311,29 @@ fn text_anchor_offset(
 }
 
 impl Span {
-    fn new(text: &str, values: Rc<ComputedValues>, dx: f64, dy: f64, depth: usize) -> Span {
+    fn new(
+        text: &str,
+        values: Rc<ComputedValues>,
+        dx: f64,
+        dy: f64,
+        depth: usize,
+        link_target: Option<String>,
+    ) -> Span {
         Span {
             values,
             text: text.to_string(),
             dx,
             dy,
             _depth: depth,
+            link_target,
         }
     }
 }
 
 impl MeasuredSpan {
     fn from_span(
+        layout_context: &LayoutContext,
         span: &Span,
-        text_writing_mode: WritingMode,
         draw_ctx: &DrawingCtx,
     ) -> MeasuredSpan {
         let values = span.values.clone();
@@ -247,8 +341,20 @@ impl MeasuredSpan {
         let view_params = draw_ctx.get_view_params();
         let params = NormalizeParams::new(&values, &view_params);
 
-        let properties = FontProperties::new(&values, text_writing_mode, &params);
-        let layout = create_pango_layout(draw_ctx, &properties, &span.text);
+        let properties = FontProperties::new(&values, &params);
+
+        let bidi_control = BidiControl::from_unicode_bidi_and_direction(
+            properties.unicode_bidi,
+            properties.direction,
+        );
+
+        let with_control_chars = wrap_with_direction_control_chars(&span.text, &bidi_control);
+        let layout = create_pango_layout(
+            draw_ctx,
+            layout_context.writing_mode,
+            &properties,
+            &with_control_chars,
+        );
         let (w, h) = layout.size();
 
         let w = f64::from(w) / f64::from(pango::SCALE);
@@ -258,7 +364,7 @@ impl MeasuredSpan {
         assert!(w >= 0.0);
         assert!(h >= 0.0);
 
-        let advance = if text_writing_mode.is_horizontal() {
+        let advance = if layout_context.writing_mode.is_horizontal() {
             (w, 0.0)
         } else {
             (0.0, w)
@@ -267,23 +373,73 @@ impl MeasuredSpan {
         MeasuredSpan {
             values,
             layout,
-            _layout_size: (w, h),
+            layout_size: (w, h),
             advance,
             dx: span.dx,
             dy: span.dy,
+            link_target: span.link_target.clone(),
         }
     }
 }
 
+// FIXME: should the pango crate provide this like PANGO_GRAVITY_IS_VERTICAL() ?
+fn gravity_is_vertical(gravity: pango::Gravity) -> bool {
+    matches!(gravity, pango::Gravity::East | pango::Gravity::West)
+}
+
+fn compute_text_box(
+    layout: &pango::Layout,
+    x: f64,
+    y: f64,
+    transform: Transform,
+    gravity: pango::Gravity,
+) -> Option<BoundingBox> {
+    #![allow(clippy::many_single_char_names)]
+
+    let (ink, _) = layout.extents();
+    if ink.width == 0 || ink.height == 0 {
+        return None;
+    }
+
+    let ink_x = f64::from(ink.x);
+    let ink_y = f64::from(ink.y);
+    let ink_width = f64::from(ink.width);
+    let ink_height = f64::from(ink.height);
+    let pango_scale = f64::from(pango::SCALE);
+
+    let (x, y, w, h) = if gravity_is_vertical(gravity) {
+        (
+            x + (ink_x - ink_height) / pango_scale,
+            y + ink_y / pango_scale,
+            ink_height / pango_scale,
+            ink_width / pango_scale,
+        )
+    } else {
+        (
+            x + ink_x / pango_scale,
+            y + ink_y / pango_scale,
+            ink_width / pango_scale,
+            ink_height / pango_scale,
+        )
+    };
+
+    let r = Rect::new(x, y, x + w, y + h);
+    let bbox = BoundingBox::new()
+        .with_transform(transform)
+        .with_rect(r)
+        .with_ink_rect(r);
+
+    Some(bbox)
+}
+
 impl PositionedSpan {
-    fn draw(
+    fn layout(
         &self,
         acquired_nodes: &mut AcquiredNodes<'_>,
-        draw_ctx: &mut DrawingCtx,
-        clipping: bool,
-    ) -> Result<BoundingBox, RenderingError> {
-        let view_params = draw_ctx.get_view_params();
-        let params = NormalizeParams::new(&self.values, &view_params);
+        draw_ctx: &DrawingCtx,
+        view_params: &ViewParams,
+    ) -> LayoutSpan {
+        let params = NormalizeParams::new(&self.values, view_params);
 
         let layout = self.layout.clone();
         let is_visible = self.values.is_visible();
@@ -291,6 +447,10 @@ impl PositionedSpan {
 
         let stroke = Stroke::new(&self.values, &params);
 
+        let gravity = layout.context().unwrap().gravity();
+
+        let bbox = compute_text_box(&layout, x, y, draw_ctx.get_transform(), gravity);
+
         let stroke_paint = self.values.stroke().0.resolve(
             acquired_nodes,
             self.values.stroke_opacity().0,
@@ -307,20 +467,24 @@ impl PositionedSpan {
             None,
         );
 
+        let paint_order = self.values.paint_order();
         let text_rendering = self.values.text_rendering();
 
-        let span = layout::TextSpan {
+        LayoutSpan {
             layout,
+            gravity,
+            bbox,
             is_visible,
             x,
             y,
+            paint_order,
             stroke,
             stroke_paint,
             fill_paint,
             text_rendering,
-        };
-
-        draw_ctx.draw_text_span(&view_params, &span, acquired_nodes, &self.values, clipping)
+            values: self.values.clone(),
+            link_target: self.link_target.clone(),
+        }
     }
 }
 
@@ -332,18 +496,27 @@ fn children_to_chunks(
     node: &Node,
     acquired_nodes: &mut AcquiredNodes<'_>,
     cascaded: &CascadedValues<'_>,
-    draw_ctx: &mut DrawingCtx,
+    draw_ctx: &DrawingCtx,
     dx: f64,
     dy: f64,
     depth: usize,
     link: Option<String>,
 ) {
+    let mut dx = dx;
+    let mut dy = dy;
+
     for child in node.children() {
         if child.is_chars() {
             let values = cascaded.get();
-            child
-                .borrow_chars()
-                .to_chunks(&child, Rc::new(values.clone()), chunks, dx, dy, depth);
+            child.borrow_chars().to_chunks(
+                &child,
+                Rc::new(values.clone()),
+                chunks,
+                dx,
+                dy,
+                depth,
+                link.clone(),
+            );
         } else {
             assert!(child.is_element());
 
@@ -364,8 +537,14 @@ fn children_to_chunks(
                 }
 
                 Element::Link(ref link) => {
-                    // TSpan::default tes all offsets to 0,
+                    // TSpan::default sets all offsets to 0,
                     // which is what we want in links.
+                    //
+                    // FIXME: This is the only place in the code where an element's method (TSpan::to_chunks)
+                    // is called with a node that is not the element itself: here, `child` is a Link, not a TSpan.
+                    //
+                    // The code works because the `tspan` is dropped immediately after calling to_chunks and no
+                    // references are retained for it.
                     let tspan = TSpan::default();
                     let cascaded = CascadedValues::new(cascaded, &child);
                     tspan.to_chunks(
@@ -389,6 +568,10 @@ fn children_to_chunks(
                 _ => (),
             }
         }
+
+        // After the first span, we don't need to carry over the parent's dx/dy.
+        dx = 0.0;
+        dy = 0.0;
     }
 }
 
@@ -462,20 +645,18 @@ impl Chars {
         dx: f64,
         dy: f64,
         depth: usize,
-    ) -> Option<Span> {
+        link_target: Option<String>,
+    ) -> Span {
         self.ensure_normalized_string(node, &*values);
 
-        if self.space_normalized.borrow().as_ref().unwrap() == "" {
-            None
-        } else {
-            Some(Span::new(
-                self.space_normalized.borrow().as_ref().unwrap(),
-                values,
-                dx,
-                dy,
-                depth,
-            ))
-        }
+        Span::new(
+            self.space_normalized.borrow().as_ref().unwrap(),
+            values,
+            dx,
+            dy,
+            depth,
+            link_target,
+        )
     }
 
     fn to_chunks(
@@ -486,13 +667,13 @@ impl Chars {
         dx: f64,
         dy: f64,
         depth: usize,
+        link_target: Option<String>,
     ) {
-        if let Some(span) = self.make_span(node, values, dx, dy, depth) {
-            let num_chunks = chunks.len();
-            assert!(num_chunks > 0);
+        let span = self.make_span(node, values, dx, dy, depth, link_target);
+        let num_chunks = chunks.len();
+        assert!(num_chunks > 0);
 
-            chunks[num_chunks - 1].spans.push(span);
-        }
+        chunks[num_chunks - 1].spans.push(span);
     }
 
     pub fn get_string(&self) -> String {
@@ -514,7 +695,7 @@ impl Text {
         node: &Node,
         acquired_nodes: &mut AcquiredNodes<'_>,
         cascaded: &CascadedValues<'_>,
-        draw_ctx: &mut DrawingCtx,
+        draw_ctx: &DrawingCtx,
         x: f64,
         y: f64,
     ) -> Vec<Chunk> {
@@ -524,7 +705,7 @@ impl Text {
         let view_params = draw_ctx.get_view_params();
         let params = NormalizeParams::new(values, &view_params);
 
-        chunks.push(Chunk::new(values, Some(x), Some(y), None));
+        chunks.push(Chunk::new(values, Some(x), Some(y)));
 
         let dx = self.dx.to_user(&params);
         let dy = self.dy.to_user(&params);
@@ -577,8 +758,6 @@ impl Draw for Text {
 
         let stacking_ctx = StackingContext::new(acquired_nodes, &elt, values.transform(), values);
 
-        let text_writing_mode = values.writing_mode();
-
         draw_ctx.with_discrete_layer(
             &stacking_ctx,
             acquired_nodes,
@@ -586,6 +765,10 @@ impl Draw for Text {
             clipping,
             None,
             &mut |an, dc| {
+                let layout_context = LayoutContext {
+                    writing_mode: values.writing_mode(),
+                };
+
                 let mut x = self.x.to_user(&params);
                 let mut y = self.y.to_user(&params);
 
@@ -593,7 +776,7 @@ impl Draw for Text {
 
                 let mut measured_chunks = Vec::new();
                 for chunk in &chunks {
-                    measured_chunks.push(MeasuredChunk::from_chunk(chunk, text_writing_mode, dc));
+                    measured_chunks.push(MeasuredChunk::from_chunk(&layout_context, chunk, dc));
                 }
 
                 let mut positioned_chunks = Vec::new();
@@ -602,9 +785,9 @@ impl Draw for Text {
                     let chunk_y = chunk.y.unwrap_or(y);
 
                     let positioned = PositionedChunk::from_measured(
+                        &layout_context,
                         chunk,
                         &view_params,
-                        text_writing_mode,
                         chunk_x,
                         chunk_y,
                     );
@@ -615,21 +798,53 @@ impl Draw for Text {
                     positioned_chunks.push(positioned);
                 }
 
-                let mut bbox = dc.empty_bbox();
+                let view_params = dc.get_view_params();
 
+                let mut layout_spans = Vec::new();
                 for chunk in &positioned_chunks {
                     for span in &chunk.spans {
-                        let span_bbox = match chunk.link.as_ref() {
-                            Some(l) if !l.is_empty() => {
-                                dc.with_link_tag(l.as_str(), &mut |dc| span.draw(an, dc, clipping))?
-                            }
-                            _ => span.draw(an, dc, clipping)?,
-                        };
-                        bbox.insert(&span_bbox);
+                        layout_spans.push(span.layout(an, dc, &view_params));
                     }
                 }
 
-                Ok(bbox)
+                let text_bbox = layout_spans.iter().fold(dc.empty_bbox(), |mut bbox, span| {
+                    if let Some(ref span_bbox) = span.bbox {
+                        bbox.insert(span_bbox);
+                    }
+
+                    bbox
+                });
+
+                let mut text_spans = Vec::new();
+                for span in layout_spans {
+                    let stroke_paint =
+                        span.stroke_paint
+                            .to_user_space(&text_bbox, &view_params, &span.values);
+                    let fill_paint =
+                        span.fill_paint
+                            .to_user_space(&text_bbox, &view_params, &span.values);
+
+                    let text_span = TextSpan {
+                        layout: span.layout,
+                        gravity: span.gravity,
+                        bbox: span.bbox,
+                        is_visible: span.is_visible,
+                        x: span.x,
+                        y: span.y,
+                        paint_order: span.paint_order,
+                        stroke: span.stroke,
+                        stroke_paint,
+                        fill_paint,
+                        text_rendering: span.text_rendering,
+                        link_target: span.link_target,
+                    };
+
+                    text_spans.push(text_span);
+                }
+
+                let text_layout = layout::Text { spans: text_spans };
+
+                dc.draw_text(&text_layout, an, clipping)
             },
         )
     }
@@ -685,7 +900,7 @@ fn extract_chars_children_to_chunks_recursively(
         if child.is_chars() {
             child
                 .borrow_chars()
-                .to_chunks(&child, values, chunks, 0.0, 0.0, depth)
+                .to_chunks(&child, values, chunks, 0.0, 0.0, depth, None)
         } else {
             extract_chars_children_to_chunks_recursively(chunks, &child, values, depth + 1)
         }
@@ -722,7 +937,7 @@ impl TSpan {
         node: &Node,
         acquired_nodes: &mut AcquiredNodes<'_>,
         cascaded: &CascadedValues<'_>,
-        draw_ctx: &mut DrawingCtx,
+        draw_ctx: &DrawingCtx,
         chunks: &mut Vec<Chunk>,
         dx: f64,
         dy: f64,
@@ -744,7 +959,7 @@ impl TSpan {
         let span_dy = dy + self.dy.to_user(&params);
 
         if x.is_some() || y.is_some() {
-            chunks.push(Chunk::new(values, x, y, link.clone()));
+            chunks.push(Chunk::new(values, x, y));
         }
 
         children_to_chunks(
@@ -835,15 +1050,6 @@ impl From<Direction> for pango::Direction {
     }
 }
 
-impl From<Direction> for pango::Alignment {
-    fn from(d: Direction) -> pango::Alignment {
-        match d {
-            Direction::Ltr => pango::Alignment::Left,
-            Direction::Rtl => pango::Alignment::Right,
-        }
-    }
-}
-
 impl From<WritingMode> for pango::Direction {
     fn from(m: WritingMode) -> pango::Direction {
         use WritingMode::*;
@@ -865,17 +1071,131 @@ impl From<WritingMode> for pango::Gravity {
     }
 }
 
-fn create_pango_layout(draw_ctx: &DrawingCtx, props: &FontProperties, text: &str) -> pango::Layout {
+/// Constants with Unicode's directional formatting characters
+///
+/// https://unicode.org/reports/tr9/#Directional_Formatting_Characters
+mod directional_formatting_characters {
+    /// Left-to-Right Embedding
+    ///
+    /// Treat the following text as embedded left-to-right.
+    pub const LRE: char = '\u{202a}';
+
+    /// Right-to-Left Embedding
+    ///
+    /// Treat the following text as embedded right-to-left.
+    pub const RLE: char = '\u{202b}';
+
+    /// Left-to-Right Override
+    ///
+    /// Force following characters to be treated as strong left-to-right characters.
+    pub const LRO: char = '\u{202d}';
+
+    /// Right-to-Left Override
+    ///
+    /// Force following characters to be treated as strong right-to-left characters.
+    pub const RLO: char = '\u{202e}';
+
+    /// Pop Directional Formatting
+    ///
+    /// End the scope of the last LRE, RLE, RLO, or LRO.
+    pub const PDF: char = '\u{202c}';
+
+    /// Left-to-Right Isolate
+    ///
+    /// Treat the following text as isolated and left-to-right.
+    pub const LRI: char = '\u{2066}';
+
+    /// Right-to-Left Isolate
+    ///
+    /// Treat the following text as isolated and right-to-left.
+    pub const RLI: char = '\u{2067}';
+
+    /// First Strong Isolate
+    ///
+    /// Treat the following text as isolated and in the direction of its first strong
+    /// directional character that is not inside a nested isolate.
+    pub const FSI: char = '\u{2068}';
+
+    /// Pop Directional Isolate
+    ///
+    /// End the scope of the last LRI, RLI, or FSI.
+    pub const PDI: char = '\u{2069}';
+}
+
+/// Unicode control characters to be inserted when `unicode-bidi` is specified.
+///
+/// The `unicode-bidi` property is used to change the embedding of a text span within
+/// another.  This struct contains slices with the control characters that must be
+/// inserted into the text stream at the span's limits so that the bidi/shaping engine
+/// will know what to do.
+struct BidiControl {
+    start: &'static [char],
+    end: &'static [char],
+}
+
+impl BidiControl {
+    /// Creates a `BidiControl` from the properties that determine it.
+    ///
+    /// See the table titled "Bidi control codes injected..." in
+    /// https://www.w3.org/TR/css-writing-modes-3/#unicode-bidi
+    #[rustfmt::skip]
+    fn from_unicode_bidi_and_direction(unicode_bidi: UnicodeBidi, direction: Direction) -> BidiControl {
+        use UnicodeBidi::*;
+        use Direction::*;
+        use directional_formatting_characters::*;
+
+        let (start, end) = match (unicode_bidi, direction) {
+            (Normal,          _)   => (&[][..],         &[][..]),
+            (Embed,           Ltr) => (&[LRE][..],      &[PDF][..]),
+            (Embed,           Rtl) => (&[RLE][..],      &[PDF][..]),
+            (Isolate,         Ltr) => (&[LRI][..],      &[PDI][..]),
+            (Isolate,         Rtl) => (&[RLI][..],      &[PDI][..]),
+            (BidiOverride,    Ltr) => (&[LRO][..],      &[PDF][..]),
+            (BidiOverride,    Rtl) => (&[RLO][..],      &[PDF][..]),
+            (IsolateOverride, Ltr) => (&[FSI, LRO][..], &[PDF, PDI][..]),
+            (IsolateOverride, Rtl) => (&[FSI, RLO][..], &[PDF, PDI][..]),
+            (Plaintext,       Ltr) => (&[FSI][..],      &[PDI][..]),
+            (Plaintext,       Rtl) => (&[FSI][..],      &[PDI][..]),
+        };
+
+        BidiControl { start, end }
+    }
+}
+
+/// Prepends and appends Unicode directional formatting characters.
+fn wrap_with_direction_control_chars(s: &str, bidi_control: &BidiControl) -> String {
+    let mut res =
+        String::with_capacity(s.len() + bidi_control.start.len() + bidi_control.end.len());
+
+    for &ch in bidi_control.start {
+        res.push(ch);
+    }
+
+    res.push_str(s);
+
+    for &ch in bidi_control.end {
+        res.push(ch);
+    }
+
+    res
+}
+
+fn create_pango_layout(
+    draw_ctx: &DrawingCtx,
+    writing_mode: WritingMode,
+    props: &FontProperties,
+    text: &str,
+) -> pango::Layout {
     let pango_context = draw_ctx.create_pango_context();
 
     if let XmlLang(Some(ref lang)) = props.xml_lang {
         pango_context.set_language(&pango::Language::from_string(lang.as_str()));
     }
 
-    pango_context.set_base_gravity(pango::Gravity::from(props.writing_mode));
+    pango_context.set_base_gravity(pango::Gravity::from(writing_mode));
 
     match (props.unicode_bidi, props.direction) {
-        (UnicodeBidi::Override, _) | (UnicodeBidi::Embed, _) => {
+        (UnicodeBidi::BidiOverride, _) | (UnicodeBidi::Embed, _) => {
             pango_context.set_base_dir(pango::Direction::from(props.direction));
         }
 
@@ -884,11 +1204,36 @@ fn create_pango_layout(draw_ctx: &DrawingCtx, props: &FontProperties, text: &str
         }
 
         (_, _) => {
-            pango_context.set_base_dir(pango::Direction::from(props.writing_mode));
+            pango_context.set_base_dir(pango::Direction::from(writing_mode));
         }
     }
 
-    let mut font_desc = pango_context.font_description().unwrap();
+    let layout = pango::Layout::new(&pango_context);
+
+    let attr_list = pango::AttrList::new();
+    add_pango_attributes(&attr_list, props, 0, text.len());
+
+    layout.set_attributes(Some(&attr_list));
+    layout.set_text(text);
+    layout.set_auto_dir(false);
+
+    layout
+}
+
+/// Adds Pango attributes, suitable for a span of text, to an `AttrList`.
+fn add_pango_attributes(
+    attr_list: &pango::AttrList,
+    props: &FontProperties,
+    start_index: usize,
+    end_index: usize,
+) {
+    let start_index = u32::try_from(start_index).expect("Pango attribute index must fit in u32");
+    let end_index = u32::try_from(end_index).expect("Pango attribute index must fit in u32");
+    assert!(start_index <= end_index);
+
+    let mut attributes = Vec::new();
+
+    let mut font_desc = pango::FontDescription::new();
     font_desc.set_family(props.font_family.as_str());
     font_desc.set_style(pango::Style::from(props.font_style));
 
@@ -901,48 +1246,43 @@ fn create_pango_layout(draw_ctx: &DrawingCtx, props: &FontProperties, text: &str
 
     font_desc.set_size(to_pango_units(props.font_size));
 
-    let layout = pango::Layout::new(&pango_context);
-    layout.set_auto_dir(false);
-    layout.set_font_description(Some(&font_desc));
+    attributes.push(pango::Attribute::new_font_desc(&font_desc));
 
-    // FIXME: For now we ignore the `line-height` property, even though we parse it.
-    // We would need to do something like this:
-    //
-    // layout.set_line_spacing(0.0); // "actually use the spacing I'll give you"
-    // layout.set_spacing(to_pango_units(???));
-    //
-    // However, Layout::set_spacing() takes an inter-line spacing (from the baseline of
-    // one line to the top of the next line), not the line height (from baseline to
-    // baseline).
-    //
-    // Maybe we need to implement layout of individual lines by hand.
-
-    let attr_list = pango::AttrList::new();
-
-    attr_list.insert(pango::Attribute::new_letter_spacing(to_pango_units(
+    attributes.push(pango::Attribute::new_letter_spacing(to_pango_units(
         props.letter_spacing,
     )));
 
+    if props.text_decoration.overline {
+        attributes.push(pango::Attribute::new_overline(pango::Overline::Single));
+    }
+
     if props.text_decoration.underline {
-        attr_list.insert(pango::Attribute::new_underline(pango::Underline::Single));
+        attributes.push(pango::Attribute::new_underline(pango::Underline::Single));
     }
 
     if props.text_decoration.strike {
-        attr_list.insert(pango::Attribute::new_strikethrough(true));
+        attributes.push(pango::Attribute::new_strikethrough(true));
     }
 
     // FIXME: Using the "smcp" OpenType feature only works for fonts that support it.  We
     // should query if the font supports small caps, and synthesize them if it doesn't.
     if props.font_variant == FontVariant::SmallCaps {
         // smcp - small capitals - https://docs.microsoft.com/en-ca/typography/opentype/spec/features_pt#smcp
-        attr_list.insert(pango::Attribute::new_font_features("'smcp' 1"));
+        attributes.push(pango::Attribute::new_font_features("'smcp' 1"));
     }
 
-    layout.set_attributes(Some(&attr_list));
-    layout.set_alignment(pango::Alignment::from(props.direction));
-    layout.set_text(text);
+    // Set the range in each attribute
 
-    layout
+    for attr in &mut attributes {
+        attr.set_start_index(start_index);
+        attr.set_end_index(end_index);
+    }
+
+    // Add the attributes to the attr_list
+
+    for attr in attributes {
+        attr_list.insert(attr);
+    }
 }
 
 #[cfg(test)]
@@ -972,18 +1312,33 @@ mod tests {
         use TextAnchor::*;
 
         assert_eq!(
-            text_anchor_offset(Start, Ltr, WritingMode::Lr, (2.0, 4.0)),
-            (0.0, 0.0)
+            text_anchor_offset(
+                Start,
+                Ltr,
+                WritingMode::Lr,
+                Rect::from_size(1.0, 2.0).translate((5.0, 6.0))
+            ),
+            (-5.0, 0.0)
         );
 
         assert_eq!(
-            text_anchor_offset(Middle, Ltr, WritingMode::Lr, (2.0, 4.0)),
-            (-1.0, 0.0)
+            text_anchor_offset(
+                Middle,
+                Ltr,
+                WritingMode::Lr,
+                Rect::from_size(1.0, 2.0).translate((5.0, 6.0))
+            ),
+            (-5.5, 0.0)
         );
 
         assert_eq!(
-            text_anchor_offset(End, Ltr, WritingMode::Lr, (2.0, 4.0)),
-            (-2.0, 0.0)
+            text_anchor_offset(
+                End,
+                Ltr,
+                WritingMode::Lr,
+                Rect::from_size(1.0, 2.0).translate((5.0, 6.0))
+            ),
+            (-6.0, 0.0)
         );
     }
 
@@ -993,16 +1348,31 @@ mod tests {
         use TextAnchor::*;
 
         assert_eq!(
-            text_anchor_offset(Start, Rtl, WritingMode::Rl, (2.0, 4.0)),
-            (0.0, 0.0)
+            text_anchor_offset(
+                Start,
+                Rtl,
+                WritingMode::Rl,
+                Rect::from_size(1.0, 2.0).translate((5.0, 6.0))
+            ),
+            (-6.0, 0.0)
         );
         assert_eq!(
-            text_anchor_offset(Middle, Rtl, WritingMode::Rl, (2.0, 4.0)),
-            (1.0, 0.0)
+            text_anchor_offset(
+                Middle,
+                Rtl,
+                WritingMode::Rl,
+                Rect::from_size(1.0, 2.0).translate((5.0, 6.0))
+            ),
+            (-5.5, 0.0)
         );
         assert_eq!(
-            text_anchor_offset(TextAnchor::End, Direction::Rtl, WritingMode::Rl, (2.0, 4.0)),
-            (2.0, 0.0)
+            text_anchor_offset(
+                TextAnchor::End,
+                Direction::Rtl,
+                WritingMode::Rl,
+                Rect::from_size(1.0, 2.0).translate((5.0, 6.0))
+            ),
+            (-5.0, 0.0)
         );
     }
 
@@ -1016,17 +1386,17 @@ mod tests {
         use TextAnchor::*;
 
         assert_eq!(
-            text_anchor_offset(Start, Ltr, WritingMode::Tb, (2.0, 4.0)),
+            text_anchor_offset(Start, Ltr, WritingMode::Tb, Rect::from_size(2.0, 4.0)),
             (0.0, 0.0)
         );
 
         assert_eq!(
-            text_anchor_offset(Middle, Ltr, WritingMode::Tb, (2.0, 4.0)),
+            text_anchor_offset(Middle, Ltr, WritingMode::Tb, Rect::from_size(2.0, 4.0)),
             (0.0, -2.0)
         );
 
         assert_eq!(
-            text_anchor_offset(End, Ltr, WritingMode::Tb, (2.0, 4.0)),
+            text_anchor_offset(End, Ltr, WritingMode::Tb, Rect::from_size(2.0, 4.0)),
             (0.0, -4.0)
         );
     }
