@@ -337,6 +337,22 @@ fn pdf_has_link() {
         .stdout(file::is_pdf().with_link("https://example.com"));
 }
 
+#[cfg(system_deps_have_cairo_pdf)]
+#[test]
+fn pdf_has_link_inside_text() {
+    let input = Path::new("tests/fixtures/cmdline/text-a-link.svg");
+    RsvgConvert::new()
+        .arg("--format=pdf")
+        .arg(input)
+        .assert()
+        .success()
+        .stdout(
+            file::is_pdf()
+                .with_link("https://example.com")
+                .and(file::is_pdf().with_link("https://another.example.com")),
+        );
+}
+
 #[cfg(system_deps_have_cairo_pdf)]
 #[test]
 fn env_source_data_epoch_controls_pdf_creation_date() {
